@@ -50,18 +50,27 @@ struct intel_mode {
 	int cpp;
 
 	drmEventContext event_context;
-	void *event_data;
+	DRI2FrameEventPtr flip_info;
 	int old_fb_id;
 	int flip_count;
+	unsigned int fe_frame;
+	unsigned int fe_tv_sec;
+	unsigned int fe_tv_usec;
 
 	struct list outputs;
 	struct list crtcs;
 };
 
+struct intel_pageflip {
+	struct intel_mode *mode;
+	Bool dispatch_me;
+};
+
 struct intel_crtc {
 	struct intel_mode *mode;
 	drmModeModeInfo kmode;
 	drmModeCrtcPtr mode_crtc;
+	int pipe;
 	dri_bo *cursor;
 	dri_bo *rotate_bo;
 	uint32_t rotate_pitch;
@@ -101,13 +110,15 @@ struct intel_output {
 static void
 intel_output_dpms(xf86OutputPtr output, int mode);
 
+static void
+intel_output_dpms_backlight(xf86OutputPtr output, int oldmode, int mode);
+
 #define BACKLIGHT_CLASS "/sys/class/backlight"
 
 /*
  * List of available kernel interfaces in priority order
  */
 static const char *backlight_interfaces[] = {
-	"intel", /* prefer our own native backlight driver */
 	"asus-laptop",
 	"eeepc",
 	"thinkpad_screen",
@@ -117,6 +128,7 @@ static const char *backlight_interfaces[] = {
 	"samsung",
 	"acpi_video1", /* finally fallback to the generic acpi drivers */
 	"acpi_video0",
+	"intel_backlight",
 	NULL,
 };
 /*
@@ -359,7 +371,6 @@ intel_crtc_apply(xf86CrtcPtr crtc)
 			       crtc->gamma_blue, crtc->gamma_size);
 #endif
 
-
 	x = crtc->x;
 	y = crtc->y;
 	fb_id = mode->fb_id;
@@ -375,9 +386,25 @@ intel_crtc_apply(xf86CrtcPtr crtc)
 		xf86DrvMsg(crtc->scrn->scrnIndex, X_ERROR,
 			   "failed to set mode: %s\n", strerror(-ret));
 		ret = FALSE;
-	} else
+	} else {
 		ret = TRUE;
 
+		/* Force DPMS to On for all outputs, which the kernel will have done
+		 * with the mode set. Also, restore the backlight level
+		 */
+		for (i = 0; i < xf86_config->num_output; i++) {
+		    xf86OutputPtr output = xf86_config->output[i];
+		    struct intel_output *intel_output;
+
+		    if (output->crtc != crtc)
+			continue;
+
+		    intel_output = output->driver_private;
+		    intel_output_dpms_backlight(output, intel_output->dpms_mode, DPMSModeOn);
+		    intel_output->dpms_mode = DPMSModeOn;
+		}
+	}
+
 	intel_set_gem_max_sizes(scrn);
 
 	if (scrn->pScreen)
@@ -424,6 +451,8 @@ intel_crtc_set_mode_major(xf86CrtcPtr crtc, DisplayModePtr mode,
 	crtc->y = y;
 	crtc->rotation = rotation;
 
+	intel_batch_submit(crtc->scrn);
+
 	mode_to_kmode(crtc->scrn, &intel_crtc->kmode, mode);
 	ret = intel_crtc_apply(crtc);
 	if (!ret) {
@@ -650,6 +679,9 @@ intel_crtc_init(ScrnInfoPtr scrn, struct intel_mode *mode, int num)
 	intel_crtc->mode = mode;
 	crtc->driver_private = intel_crtc;
 
+	intel_crtc->pipe = drm_intel_get_pipe_from_crtc_id(intel->bufmgr,
+							   crtc_id(intel_crtc));
+
 	intel_crtc->cursor = drm_intel_bo_alloc(intel->bufmgr, "ARGB cursor",
 						HWCURSOR_SIZE_ARGB,
 						GTT_PAGE_SIZE);
@@ -1161,7 +1193,11 @@ intel_output_set_property(xf86OutputPtr output, Atom property,
 		}
 	}
 
-	return FALSE;
+	/* We didn't recognise this property, just report success in order
+	 * to allow the set to continue, otherwise we break setting of
+	 * common properties like EDID.
+	 */
+	return TRUE;
 }
 
 static Bool
@@ -1327,6 +1363,8 @@ intel_xf86crtc_resize(ScrnInfoPtr scrn, int width, int height)
 	if (scrn->virtualX == width && scrn->virtualY == height)
 		return TRUE;
 
+	intel_batch_submit(scrn);
+
 	old_width = scrn->virtualX;
 	old_height = scrn->virtualY;
 	old_pitch = scrn->displayWidth;
@@ -1390,13 +1428,14 @@ fail:
 Bool
 intel_do_pageflip(intel_screen_private *intel,
 		  dri_bo *new_front,
-		  void *data)
+		  DRI2FrameEventPtr flip_info, int ref_crtc_hw_id)
 {
 	ScrnInfoPtr scrn = intel->scrn;
 	xf86CrtcConfigPtr config = XF86_CRTC_CONFIG_PTR(scrn);
 	struct intel_crtc *crtc = config->crtc[0]->driver_private;
 	struct intel_mode *mode = crtc->mode;
 	unsigned int pitch = scrn->displayWidth * intel->cpp;
+	struct intel_pageflip *flip;
 	int i, old_fb_id;
 
 	/*
@@ -1417,18 +1456,39 @@ intel_do_pageflip(intel_screen_private *intel,
 	 * Also, flips queued on disabled or incorrectly configured displays
 	 * may never complete; this is a configuration error.
 	 */
+	mode->fe_frame = 0;
+	mode->fe_tv_sec = 0;
+	mode->fe_tv_usec = 0;
+
 	for (i = 0; i < config->num_crtc; i++) {
 		if (!config->crtc[i]->enabled)
 			continue;
 
-		mode->event_data = data;
+		mode->flip_info = flip_info;
 		mode->flip_count++;
+
+		crtc = config->crtc[i]->driver_private;
+
+		flip = calloc(1, sizeof(struct intel_pageflip));
+		if (flip == NULL) {
+			xf86DrvMsg(scrn->scrnIndex, X_WARNING,
+				   "flip queue: carrier alloc failed.\n");
+			goto error_undo;
+		}
+
+		/* Only the reference crtc will finally deliver its page flip
+		 * completion event. All other crtc's events will be discarded.
+		 */
+		flip->dispatch_me = (intel_crtc_to_pipe(crtc->crtc) == ref_crtc_hw_id);
+		flip->mode = mode;
+
 		if (drmModePageFlip(mode->fd,
-				    crtc_id(config->crtc[i]->driver_private),
+				    crtc_id(crtc),
 				    mode->fb_id,
-				    DRM_MODE_PAGE_FLIP_EVENT, mode)) {
+				    DRM_MODE_PAGE_FLIP_EVENT, flip)) {
 			xf86DrvMsg(scrn->scrnIndex, X_WARNING,
 				   "flip queue failed: %s\n", strerror(errno));
+			free(flip);
 			goto error_undo;
 		}
 	}
@@ -1452,28 +1512,41 @@ static const xf86CrtcConfigFuncsRec intel_xf86crtc_config_funcs = {
 
 static void
 intel_vblank_handler(int fd, unsigned int frame, unsigned int tv_sec,
-		       unsigned int tv_usec, void *event_data)
+		       unsigned int tv_usec, void *event)
 {
-	I830DRI2FrameEventHandler(frame, tv_sec, tv_usec, event_data);
+	I830DRI2FrameEventHandler(frame, tv_sec, tv_usec, event);
 }
 
 static void
 intel_page_flip_handler(int fd, unsigned int frame, unsigned int tv_sec,
 			  unsigned int tv_usec, void *event_data)
 {
-	struct intel_mode *mode = event_data;
+	struct intel_pageflip *flip = event_data;
+	struct intel_mode *mode = flip->mode;
 
+	/* Is this the event whose info shall be delivered to higher level? */
+	if (flip->dispatch_me) {
+		/* Yes: Cache msc, ust for later delivery. */
+		mode->fe_frame = frame;
+		mode->fe_tv_sec = tv_sec;
+		mode->fe_tv_usec = tv_usec;
+	}
+	free(flip);
 
+	/* Last crtc completed flip? */
 	mode->flip_count--;
 	if (mode->flip_count > 0)
 		return;
 
+	/* Release framebuffer */
 	drmModeRmFB(mode->fd, mode->old_fb_id);
 
-	if (mode->event_data == NULL)
+	if (mode->flip_info == NULL)
 		return;
 
-	I830DRI2FlipEventHandler(frame, tv_sec, tv_usec, mode->event_data);
+	/* Deliver cached msc, ust from reference crtc to flip event handler */
+	I830DRI2FlipEventHandler(mode->fe_frame, mode->fe_tv_sec,
+				 mode->fe_tv_usec, mode->flip_info);
 }
 
 static void
@@ -1602,16 +1675,15 @@ intel_mode_fini(intel_screen_private *intel)
 	intel->modes = NULL;
 }
 
-int
-intel_get_pipe_from_crtc_id(drm_intel_bufmgr *bufmgr, xf86CrtcPtr crtc)
-{
-	return drm_intel_get_pipe_from_crtc_id(bufmgr,
-					      	crtc_id(crtc->driver_private));
-}
-
 /* for the mode overlay */
 int
 intel_crtc_id(xf86CrtcPtr crtc)
 {
 	return crtc_id(crtc->driver_private);
 }
+
+int intel_crtc_to_pipe(xf86CrtcPtr crtc)
+{
+	struct intel_crtc *intel_crtc = crtc->driver_private;
+	return intel_crtc->pipe;
+}
